(***********************************************
 *                                             *
 *              CS520 Fall 2006                *
 *           Programming Assignment            *
 *         Computer Science Department         *
 *             Boston University               *
 *                                             *
 ***********************************************)

open Format
open Support.Error
open Support.Pervasive

(* Abstract Syntax of types *)

type varrep = {name: int; link: stp option ref}

and stp =
    TpBase of string
  | TpFun of stp * stp
  | TpTup of stp list 
  | TpVar of varrep (* type variable *)
  | TpUni of int (* universal type variable *)  
          
(* Abstract Syntax Tree (AST) definitions *)
type ttm =
    TtmBool of bool
  | TtmInt of int
  | TtmStr of string
  | TtmVar of string
  | TtmIf of ttm * ttm * ttm
  | TtmOp of string * (ttm list)
  | TtmLam of string * stp option * ttm
  | TtmApp of ttm * ttm
  | TtmLet of string * ttm * ttm
  | TtmLetrec of string * stp option * ttm * ttm
  | TtmTup of ttm list
  | TtmPro of ttm * int
  | TtmFix of ttm
  | TtmAsc of ttm * stp

let rec tp2str tp = 
  match tp with
    TpBase s -> "TpBase \"" ^ s ^ "\""
  | TpFun (tp1, tp2) -> "TpFun (" ^ (tp2str tp1) ^ ", " ^ (tp2str tp2) ^ ")"
  | TpTup (tps) -> "TpTup [" ^ (tps2str tps) ^ "]"
  | TpVar ({name = n; link = l}) -> "TpVar {name = " ^ (string_of_int n) ^ "; link = " ^ (tpo2str !l) ^ "}"
  | TpUni i -> "TpUni " ^ (string_of_int i)
  
and tps2str tps = 
  match tps with
    [] -> ""
  | tp::tps -> (tp2str tp) ^ "," ^ (tps2str tps)
  
and tpo2str tpo =
  match tpo with
    Some tp -> tp2str tp
  | _ -> ""  

let rec indent i = if i = 0 then "" else " " ^ (indent (i-1))

let rec tm2str0 i t = 
  match t with
    TtmVar v -> (indent i) ^ "TtmVar \"" ^ v ^ "\""
  | TtmLam (v, tpo, t') ->
      (indent i) ^ "TtmLam (\"" ^ v ^ "\", " ^ (tpo2str tpo) ^ ",\n" 
       ^ (tm2str0 (i+8) t') ^ ")"
  | TtmApp (t1, t2) ->
       (indent i) ^ "TtmApp (\n" 
                  ^ (tm2str0 (i+8) t1) ^ ",\n" 
                  ^ (tm2str0 (i+8) t2) ^ ")"
  | TtmBool b -> (indent i) ^ "TtmBool " ^ (if b then "\"true\"" else "\"false\"")
  | TtmStr s -> (indent i) ^ "TtmStr \"" ^ s ^ "\""
  | TtmInt j -> ((indent i) ^ "TtmInt " ^ (string_of_int j))
  | TtmOp (op, args) -> 
       (indent i) ^ "TtmOp (\"" ^ op ^ "\",\n" ^ 
       (indent (i+7)) ^ "[\n" ^ (tms2str0 (i+8) args) ^ "])"
  | TtmIf (t1, t2, t3) -> (indent i) ^ "TtmIf (\n" 
                                     ^ (tm2str0 (i+7) t1) ^ ",\n" 
                                     ^ (tm2str0 (i+7) t2) ^ ",\n"
                                     ^ (tm2str0 (i+7) t3) ^ ")"
  | TtmLet (v, t1, t2) -> (indent i) ^ "TtmLet (\"" ^ v ^ "\",\n" 
                                     ^ (tm2str0 (i+8) t1) ^ ",\n"
                                     ^ (tm2str0 (i+8) t2) ^ ")"
  | TtmLetrec (v, tpo, t1, t2) -> (indent i) ^ "TtmLetrec (\"" ^ v ^ "\"," ^ (tpo2str tpo) ^ ",\n" 
                                             ^ (tm2str0 (i+11) t1) ^ ",\n"
                                             ^ (tm2str0 (i+11) t2) ^ ")"
  | TtmTup tms -> (indent i) ^ "TtmTup [\n" 
                             ^ (tms2str0 (i+8) tms) ^ "]"
  | TtmPro (t, j) -> (indent i) ^ "TtmPro (\n" ^ (tm2str0 (i+8) t) ^ ", " 
                     ^ (string_of_int j) ^ ")"
  | TtmFix (t) -> (indent i) ^ "TtmFix (\n" 
                             ^ (tm2str0 (i+8) t) ^ ")"
  | TtmAsc (tm, tp) -> (indent i) ^ "TtmAsc (\n" 
                                  ^ (tm2str0 (i+8) tm) ^ ", " ^ (tp2str tp) ^ ")"

and tms2str0 i tms = 
  match tms with
    [] -> (indent (i-1)) ^ ""
  | tm::tms -> (tm2str0 i tm) ^ ",\n" ^ (tms2str0 i tms)

let rec prettytm2str t = tm2str0 0 t

let printtm t = pr ((prettytm2str t) ^ "\n");










